@@ -19,6 +19,12 @@ static int have_buffer(out123_handle *ao)
 
 #include "debug.h"
 
+/* An output that is live and does not deal with pausing itself.
+   The device needs to be closed if we stop feeding. */
+#define SENSITIVE_OUTPUT(ao) \
+  (    (ao)->propflags & OUT123_PROP_LIVE \
+  && !((ao)->propflags & OUT123_PROP_PERSISTENT) )
+
 static const char *default_name = "out123";
 
 static int modverbose(out123_handle *ao)
@@ -50,6 +56,19 @@ static void out123_clear_module(out123_handle *ao)
 	ao->propflags = OUT123_PROP_LIVE;
 }
 
+/* Ensure that real name is not leaked, needs to be freed before any call to
+   ao->open(ao). One might free it on closing already, but it might be sensible
+   to keep it around, might still be the same after re-opening. */
+static int aoopen(out123_handle *ao)
+{
+	if(ao->realname)
+	{
+		free(ao->realname);
+		ao->realname = NULL;
+	}
+	return ao->open(ao);
+}
+
 out123_handle* attribute_align_arg out123_new(void)
 {
 	out123_handle* ao = malloc( sizeof( out123_handle ) );
@@ -430,6 +449,7 @@ void attribute_align_arg out123_close(out123_handle *ao)
 	ao->errcode = 0;
 
 	out123_drain(ao);
+	out123_stop(ao);
 
 #ifndef NOXFERMEM
 	if(have_buffer(ao))
@@ -437,17 +457,6 @@ void attribute_align_arg out123_close(out123_handle *ao)
 	else
 #endif
 	{
-		/* Only close output device if one is opened. Module code might not be resilient
-		   about that. */
-		if(ao->state > play_stopped && ao->close)
-		{
-			int ret;
-			if((ret=ao->close(ao)))
-			{
-				if(!AOQUIET)
-					error1("ao->close() returned %i", ret);
-			}
-		}
 		if(ao->deinit)
 			ao->deinit(ao);
 		if(ao->module)
@@ -505,7 +514,7 @@ out123_start(out123_handle *ao, long rate, int channels, int encoding)
 	else
 #endif
 	{
-		if(ao->open(ao) < 0)
+		if(aoopen(ao) < 0)
 			return out123_seterr(ao, OUT123_DEV_OPEN);
 		ao->state = play_live;
 		return OUT123_OK;
@@ -523,8 +532,7 @@ void attribute_align_arg out123_pause(out123_handle *ao)
 #endif
 		{
 			/* Close live devices to avoid underruns. */
-			if(    ao->propflags & OUT123_PROP_LIVE
-			  && !(ao->propflags & OUT123_PROP_PERSISTENT)
+			if( SENSITIVE_OUTPUT(ao)
 			  && ao->close && ao->close(ao) && !AOQUIET )
 				error("trouble closing device");
 		}
@@ -542,9 +550,7 @@ void attribute_align_arg out123_continue(out123_handle *ao)
 		else
 #endif
 		/* Re-open live devices to avoid underruns. */
-		if(    ao->propflags & OUT123_PROP_LIVE
-		  && !(ao->propflags & OUT123_PROP_PERSISTENT)
-		  && ao->open(ao) < 0)
+		if(SENSITIVE_OUTPUT(ao) && aoopen(ao) < 0)
 		{
 			/* Will be overwritten by following out123_play() ... */
 			ao->errcode = OUT123_DEV_OPEN;
@@ -569,8 +575,12 @@ void attribute_align_arg out123_stop(out123_handle *ao)
 		buffer_stop(ao);
 	else
 #endif
-	if(ao->close && ao->close(ao) && !AOQUIET)
-		error("trouble closing device");
+	if(   ao->state == play_live
+	  || (ao->state == play_paused && !SENSITIVE_OUTPUT(ao)) )
+	{
+		if(ao->close && ao->close(ao) && !AOQUIET)
+			error("trouble closing device");
+	}
 	ao->state = play_stopped;
 }
 
@@ -828,7 +838,7 @@ static void check_output_module( out123_handle *ao
 	if(result == 0)
 	{ /* Try to open the device. I'm only interested in actually working modules. */
 		ao->format = -1;
-		result = ao->open(ao);
+		result = aoopen(ao);
 		debug1("ao->open() = %i", result);
 		if(result >= 0)
 			ao->close(ao);
@@ -961,7 +971,7 @@ out123_encodings(out123_handle *ao, long rate, int channels)
 		   They possibly set a sample rate and channel count they like best.
 		   We should add API to retrieve those defaults, too. */
 		ao->format   = -1;
-		if(ao->open(ao) >= 0)
+		if(aoopen(ao) >= 0)
 		{
 			/* Need to reset those since the choose-your-format open
 			   call might have changed them. */
@@ -1018,7 +1028,7 @@ out123_formats( out123_handle *ao, const long *rates, int ratecount
 		ao->format   = -1;
 		ao->rate     = -1;
 		ao->channels = -1;
-		if(ao->open(ao) >= 0)
+		if(aoopen(ao) >= 0)
 		{
 			struct mpg123_fmt *fmts;
 			int ri, ch;
